//
// BNPD.CPP
//
// Sample Battle.net Proxy Server.
//
// By Scott Coleman (9/30/98)
//
//   This is a reference implementation of a proxy server which will
// allow a machine located behind a firewall, NAT, or IP
// Masquerading gateway to play games on Battle.net. This is probably not
// the fastest nor the most robust implementation possible; rather, it is
// intended to act as a guide so that those who are in the business of
// developing network proxy agents can integrate Battle.net support into
// their products. Although we welcome your comments and bug reports,
// Blizzard Entertainment cannot offer support for this software.
//
//

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <malloc.h>
#include <ctype.h>
#include <signal.h>

#ifdef WIN32

#define  STRICT
#include <windows.h>
#include <winsock.h>

#else

#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_udp.h>
#include <netinet/ip_tcp.h>
#include <netinet/ip_icmp.h>
#include <netinet/protocols.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <errno.h>

#define closesocket(s) close(s)

#endif

#include "bnpd.h"
#pragma hdrstop

#define MAX_ARGS        20

static int Cleanup();
int ParseResponseFile(char *lpcszFileName);

static u_long sguListenAddr           = INADDR_ANY;
static int sgsdUnifiedSock            = -1;     // master listening socket
static int sgsdUdpSock                = -1;     // proxy UDP socket
static struct sockaddr_in sgGameDataAddr;       // game data address on client
static FILE *sgpLog                   =  0;
static PROXYCONNECTIONPTR sgpConnHead =  0;


//============================================================================
int GetErrorCode() {

#ifdef WIN32
  return GetLastError();
#else
  return errno;
#endif
}

//============================================================================
int StrIsEmpty(const char *lpcszStr) {

  if (!lpcszStr)
    return TRUE;

  while (*lpcszStr) {
    if (!isspace(*lpcszStr))
      return FALSE;
    lpcszStr++;
  }

  return TRUE;
}

//============================================================================
char *StrStrip(char *lpszStr) {
  int i, j;
  for (i=0, j=0; lpszStr[i]; ++i)
    if ((lpszStr[i] != '\r') && (lpszStr[i] != '\n')) {
      if (i != j)
        lpszStr[j] = lpszStr[i];
      j++;
    }
  lpszStr[j] = '\0';
  return lpszStr;
}


//=============================================================================
int LogOpen(char *lpszFileName) {

  if (sgpLog) {
    LogWrite("Attempt to re-open log file.");
    return FALSE;
  }

  sgpLog = fopen(lpszFileName, "ab");
  if (!sgpLog)
    return FALSE;

  return TRUE;
}


//=============================================================================
#ifdef WIN32
int __cdecl LogWrite(char *lpszFmt, ...) {
#else
int LogWrite(char *lpszFmt, ...) {
#endif
  va_list argptr;
  char szTimeBuf[26];
  char szOutStr[1024];

  if (!sgpLog)
    return FALSE;

  va_start(argptr, lpszFmt);
  vsprintf(szOutStr, lpszFmt, argptr);
  va_end(argptr);

  // get current time
  time_t currtime = time(0);
  strcpy(szTimeBuf, ctime(&currtime));

  StrStrip(szTimeBuf);
  StrStrip(szOutStr);

  fprintf(sgpLog, "%s %s\r\n", szTimeBuf, szOutStr);
  fflush(sgpLog);
  return TRUE;
}


//=============================================================================
void LogClose() {

  if (sgpLog)
    fclose(sgpLog);
}


//============================================================================
u_long Resolve(const char *lpcszAddrStr) {
  struct hostent *hp;
  char *p = (char *)lpcszAddrStr;
  u_long uAddr;

  while (*p && (isdigit(*p) || *p == '.'))
    p++;

  if (!*p) {
    return (inet_addr(lpcszAddrStr));
  }

  hp = gethostbyname(lpcszAddrStr);
  if (!hp)
    return FALSE;

  memcpy((char *)&uAddr, hp->h_addr, sizeof(u_long));
  return uAddr;

}


//============================================================================
int ParseArgs(int argc, char *argv[]) {

  int i=0;
  while (i < argc) {
    if (argv[i][0] == '@') {
      ParseResponseFile(&argv[i][1]);
    }
    else if (argv[i][0] == '-') {
      switch(tolower(argv[i][1])) {
        case 'l':
          sguListenAddr = Resolve(&argv[i][2]);
        break;

        default:
          LogWrite("Unrecognized command line switch %c\n", argv[i][1]);
        break;
      }
    }
    i++;
  }
  return TRUE;
}


//===========================================================================
int ParseResponseFile(char *lpcszFileName) {
  FILE *f;
  int argc=0;
  char *argv[MAX_ARGS];

  f = fopen(lpcszFileName, "rt");
  if (!f) {
    LogWrite("Can't open response file %s!\n", lpcszFileName);
    return FALSE;
  }

  char szLine[256];
  while (!feof(f)) {
    if (!fgets(szLine, 255, f))
      break;
    char *p = strtok(szLine, " \t\r\n");
    while (p) {
      argv[argc] = (char *)malloc(strlen(p)+1);
      if (!argv[argc])
        break;
      strcpy(argv[argc], p);
      argc++;
      p = strtok(NULL, " \t\r\n");
    }
  }

  fclose(f);
  int bResult = ParseArgs(argc, argv);

  for (int i=0; i<argc; ++i)
    free (argv[i]);

  return bResult;
}


//=============================================================================
int ServerConnect(u_short uServerPort) {

  FILE *f = fopen("server.lst", "rt");
  if (!f)
    return -1;

  // build the list of connecting sockets
  int s = -1;
  CONNSOCKRECPTR pHead = 0;
  while (!feof(f)) {
    char szLine[256] = "";
    if (!fgets(szLine,255,f))
      break;

    StrStrip(szLine);
    if ((szLine[0] == ';') || StrIsEmpty(szLine))
      continue;

    // add a new connection attempt record
    CONNSOCKRECPTR pRec = (CONNSOCKRECPTR)calloc(sizeof(CONNSOCKREC),1);
    if (!pRec)
      break;
    pRec->sdSock = -1;
    pRec->nState = CS_DISCONNECTED;
    pRec->Addr.sin_family = AF_INET;
    if ((pRec->Addr.sin_addr.s_addr = Resolve(szLine)) != 0) {
      pRec->Addr.sin_port = htons(uServerPort);
      pRec->pNext = pHead;
      pHead = pRec;
    }
  }
  fclose(f);

  // now try to connect
  do {
    // traverse the list, opening connections and building the fd_set
    int nHiSock = 0;
    fd_set fds;
    FD_ZERO(&fds);
    CONNSOCKRECPTR pCurr = pHead;
    while (pCurr) {
      switch (pCurr->nState) {
        case CS_DISCONNECTED:
          pCurr->sdSock = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
          if (pCurr->sdSock == -1)
            break;

          // disable Nagle buffering
          u_long off = 0;
          setsockopt(pCurr->sdSock,
                     IPPROTO_TCP,
                     TCP_NODELAY,
                     (const char *)&off,
                     sizeof(off));

          // turn on keepalives
          u_long on=1;
          setsockopt(pCurr->sdSock,
                     SOL_SOCKET,
                     SO_KEEPALIVE,
                     (const char *)&on,
                     sizeof(on));

          // change to non-blocking I/O
#ifdef WIN32
          ioctlsocket(pCurr->sdSock, FIONBIO, &on);
#else
          fcntl(pCurr->sdSock, FNDELAY, &on);
#endif

          // begin the connection process
          if (connect(pCurr->sdSock,
                      (struct sockaddr *)&pCurr->Addr,
                      sizeof(pCurr->Addr)) < 0) {
#ifdef WIN32
            if (GetErrorCode() != WSAEWOULDBLOCK) {
#else
            if (GetErrorCode() != EWOULDBLOCK) {
#endif
              break;
            }
          }
          pCurr->nState = CS_CONNECTING;
        break;
      }

      // add connecting sockets to the fd_set
      if (pCurr->nState == CS_CONNECTING) {
        FD_SET(pCurr->sdSock,&fds);
        if (pCurr->sdSock > nHiSock)
          nHiSock = pCurr->sdSock;
      }
      pCurr = pCurr->pNext;
    }

    // call select() with a reasonable timeout
    struct timeval tv;
    tv.tv_sec = 5;
    tv.tv_usec = 0;
    int n = select(nHiSock+1, 0, &fds, 0, &tv);

    // check the results
    if (n < 0) {
      break;
    }

    // traverse the list again, checking active sockets
    pCurr = pHead;
    while (pCurr && (n > 0)) {
      if (FD_ISSET(pCurr->sdSock,&fds)) {
        if (s == -1) {
          // we have a winner!
          s = pCurr->sdSock;
        }
        n--;
      }
      pCurr = pCurr->pNext;
    }
  } while (s == -1);

  // close unused sockets and destroy the list
  CONNSOCKRECPTR pCurr = pHead;
  while (pCurr) {
    if (pCurr->sdSock != s) {
      closesocket(pCurr->sdSock);
    }
    pHead = pCurr->pNext;
    free(pCurr);
    pCurr = pHead;
  }
  return s;
}


//============================================================================
int AddConnection (int sdNewSock, struct sockaddr_in *cliaddr) {

  // allocate a new connection record
  PROXYCONNECTIONPTR pConn =
    (PROXYCONNECTIONPTR)malloc(sizeof(PROXYCONNECTION));
  if (!pConn) {
    return FALSE;
  }

  pConn->nType         = CONNTYPE_NONE;
  pConn->uFlags        = 0;
  pConn->sdClientSock  = sdNewSock;
  pConn->nClientBufPos = 0;
  pConn->nClientBufLen = 0;
  pConn->nServerBufPos = 0;
  pConn->nServerBufLen = 0;

  // set up the connection to the client
  {
    // disable Nagle buffering
    u_long off = 0;
    setsockopt(pConn->sdClientSock,
               IPPROTO_TCP,
               TCP_NODELAY,
               (const char *)&off,
               sizeof(off));

    // turn on keepalives
    u_long on=1;
    setsockopt(pConn->sdClientSock,
               SOL_SOCKET,
               SO_KEEPALIVE,
               (const char *)&on,
               sizeof(on));

    // change to non-blocking I/O
#ifdef WIN32
    ioctlsocket(pConn->sdClientSock, FIONBIO, &on);
#else
    fcntl(pConn->sdClientSock, FNDELAY, &on);
#endif
  }

  // connect to the server
  if ((pConn->sdServerSock = ServerConnect(UNIFIEDPORT)) == -1) {
    goto failout;
  }

  // add this connection to the session
  pConn->pNext = sgpConnHead;
  sgpConnHead = pConn;

  return TRUE;

failout:
  if (pConn)
    free(pConn);

  return FALSE;
}


//============================================================================
int KillConnection (PROXYCONNECTIONPTR pConn) {

  closesocket(pConn->sdClientSock);
  pConn->sdClientSock = -1;
  closesocket(pConn->sdServerSock);
  pConn->sdServerSock = -1;
  pConn->uFlags |= CF_DESTROY;

  return TRUE;
}


//=============================================================================
int SendMessage(int sock, unsigned char msgid, char *data, int datalen) {
  MSGHDR resphdr = { 0xff, msgid, sizeof(MSGHDR)+datalen };

  if (send(sock, (char *)&resphdr, sizeof(MSGHDR), 0) <= 0) {
    return FALSE;
  }

  if (datalen)
    if (send(sock, data, datalen, 0) <= 0) {
      return FALSE;
    }

  return TRUE;
}


//============================================================================
int SaveServerList(const char *pBuf, int nLen) {

  if (!nLen)
    return TRUE;

  char *pTmpBuf = (char *)malloc(nLen);
  if (!pTmpBuf)
    return FALSE;
  memcpy(pTmpBuf, pBuf, nLen);

  FILE *lst = fopen("server.lst", "wt");
  if (!lst)
    goto failout;

  {
    char *p = strtok(pTmpBuf, ";\r\n");
    while (p) {
      fprintf(lst, "%s\n", p);
      p = strtok(NULL, ";\r\n");
    }
  }

  fclose(lst);

failout:
  free(pTmpBuf);
  return TRUE;
}


//============================================================================
int SnagServerMessage(PROXYCONNECTIONPTR pConn,
                      int nMsgType,
                      char *msgData,
                      int nMsgLen) {
  int fSnagged = FALSE;

  switch (nMsgType) {
    case SID_SERVERLIST:
      SaveServerList(msgData+sizeof(u_long), nMsgLen-sizeof(u_long));
      fSnagged = TRUE;
    break;
  }

  return fSnagged;
}


//============================================================================
int SnagClientMessage (PROXYCONNECTIONPTR pConn,
                       int nMsgType,
                       char *msgData,
                       int nMsgLen) {
  int fSnagged = FALSE;
  struct sockaddr_in proxyaddr;
  int addrlen = sizeof(proxyaddr);

  switch (nMsgType) {
    case SID_STARTADVEX2:
      // stuff in the proxy port
      getsockname(sgsdUdpSock, (struct sockaddr *)&proxyaddr, &addrlen);
      // NOTE: PORT IS INTEL BYTE ORDER!
      *(u_short *)(msgData+PORTOFFSET) = ntohs(proxyaddr.sin_port);

      fSnagged = SendMessage(pConn->sdServerSock,
                             SID_STARTADVEX2,
                             msgData,
                             nMsgLen);
    break;

    case SID_GAMEDATAADDRESS:
      // save actual client address and port
      struct sockaddr_in *pGameDataAddr = (struct sockaddr_in *)msgData;
      memcpy(&sgGameDataAddr, pGameDataAddr, nMsgLen);

      // stuff in the proxy address and port
      getsockname(sgsdUdpSock, (struct sockaddr *)&proxyaddr, &addrlen);
      pGameDataAddr->sin_family = proxyaddr.sin_family;
      pGameDataAddr->sin_addr.s_addr = proxyaddr.sin_addr.s_addr;
      pGameDataAddr->sin_port = proxyaddr.sin_port;

      fSnagged = SendMessage(pConn->sdServerSock,
                             SID_GAMEDATAADDRESS,
                             msgData,
                             nMsgLen);
    break;
  }
  return fSnagged;
}


//============================================================================
int ProcessUdpData () {
  struct sockaddr_in peer;
  char buf[1500];

  int peerlen = sizeof(peer);
  int n = recvfrom(sgsdUdpSock, buf, 1500, 0,
                   (struct sockaddr *)&peer, &peerlen);

  if (n <= 0)
    return FALSE;

  // forward the packet to the client
  if (sendto(sgsdUdpSock, buf, n, 0,
         (struct sockaddr *)&sgGameDataAddr,
         sizeof(sgGameDataAddr)) <= 0) {
    LogWrite("Error %d forwarding UDP packet.", GetErrorCode());
    return FALSE;
  }

  return TRUE;
}


//============================================================================
int ProcessClientData (PROXYCONNECTIONPTR pConn) {

  // read data from the socket
  int nBytesToRead = STAGEBUFSIZE-pConn->nClientBufLen-pConn->nClientBufPos;
  int n = recv(pConn->sdClientSock,
               pConn->ClientStageBuf+pConn->nClientBufPos+pConn->nClientBufLen,
               nBytesToRead,
               0);
  if (n <= 0) {
    KillConnection(pConn);
    return TRUE;
  }

  pConn->nClientBufLen += n;

  // set connection type if necessary
  if ((pConn->nType == CONNTYPE_NONE) && (pConn->nClientBufLen > 0)){
    pConn->nType = pConn->ClientStageBuf[pConn->nClientBufPos];

    // check connection type
    if (pConn->nType == CONNTYPE_CLIENT) {
      // set up this session's UDP socket
      if (sgsdUdpSock == -1) {
        sgsdUdpSock = socket(PF_INET,SOCK_DGRAM,IPPROTO_UDP);

        struct sockaddr_in udpname;
        memset(&udpname, '\0', sizeof(udpname));
        int namelen = sizeof(udpname);
        if (getsockname(pConn->sdServerSock,
                        (struct sockaddr *)&udpname,
                        &namelen) < 0) {
          LogWrite("Error %d getting bound address.", GetErrorCode());
          return FALSE;
        }

        // bind to the same interface we used for our server connection;
        // allocate an unused port
        udpname.sin_port = 0;
        if (bind(sgsdUdpSock, (struct sockaddr *)&udpname, namelen) < 0) {
          return FALSE;
        }

        namelen = sizeof(udpname);
        if (getsockname(sgsdUdpSock,
                        (struct sockaddr *)&udpname,
                        &namelen) < 0) {
        }
      }
      else {
        // duplicate session attempt
        LogWrite("Rejecting second session.");
        KillConnection(pConn);
        return FALSE;
      }
    }

    // forward connection type to the server
    n = send(pConn->sdServerSock,
             pConn->ClientStageBuf+pConn->nClientBufPos,
             1,
             0);

    pConn->nClientBufPos++;
    pConn->nClientBufLen--;
  }

  if (pConn->nType == CONNTYPE_CLIENT) {

    // parse and process the messages
    while (pConn->nClientBufLen) {
      // make sure we have a complete message
      MSGHDRPTR m = (MSGHDRPTR)(pConn->ClientStageBuf+pConn->nClientBufPos);
      if (m->signature != 0xff) {
        break;
      }
      if ((pConn->nClientBufLen < sizeof(MSGHDR)) ||
          (m->bytes > pConn->nClientBufLen)) {
        break;
      }

      // intercept key messages here
      if (!SnagClientMessage(pConn,
                             m->msgtype,
                             pConn->ClientStageBuf+
                             pConn->nClientBufPos+
                             sizeof(MSGHDR),
                             m->bytes-sizeof(MSGHDR))) {
        // forward this message
        n = send(pConn->sdServerSock,
                 pConn->ClientStageBuf+pConn->nClientBufPos,
                 m->bytes,
                 0);
        if (n <= 0) {
          KillConnection(pConn);
          return TRUE;
        }
      }
      pConn->nClientBufLen -= m->bytes;
      pConn->nClientBufPos += m->bytes;
    }
  }
  else {
    // forward all available data
    n = send(pConn->sdServerSock,
             pConn->ClientStageBuf+pConn->nClientBufPos,
             pConn->nClientBufLen,
             0);

    if (n <= 0) {
      KillConnection(pConn);
      return TRUE;
    }
    pConn->nClientBufLen -= n;
    pConn->nClientBufPos += n;
  }

  if (pConn->nClientBufLen == 0) {
    pConn->nClientBufPos = 0;
  }
  return FALSE;
}


//============================================================================
int ProcessServerData (PROXYCONNECTIONPTR pConn) {

  // read data from the socket
  int nBytesToRead = STAGEBUFSIZE-pConn->nServerBufLen-pConn->nServerBufPos;
  int n = recv(pConn->sdServerSock,
               pConn->ServerStageBuf+pConn->nServerBufPos+pConn->nServerBufLen,
               nBytesToRead,
               0);
  if (n <= 0) {
    KillConnection(pConn);
    return TRUE;
  }

  pConn->nServerBufLen += n;

  if (pConn->nType == CONNTYPE_CLIENT) {
    // parse and process the messages
    while (pConn->nServerBufLen) {
      // make sure we have a complete message
      MSGHDRPTR m = (MSGHDRPTR)(pConn->ServerStageBuf+pConn->nServerBufPos);
      if (m->signature != 0xff) {
        break;
      }
      if ((pConn->nServerBufLen < sizeof(MSGHDR)) ||
          (m->bytes > pConn->nServerBufLen)) {
        break;
      }

      // intercept key messages here
      if (!SnagServerMessage(pConn,
                             m->msgtype,
                             pConn->ServerStageBuf+
                             pConn->nServerBufPos+
                             sizeof(MSGHDR),
                             m->bytes-sizeof(MSGHDR))) {
        // forward this message
        n = send(pConn->sdClientSock,
                 pConn->ServerStageBuf+pConn->nServerBufPos,
                 m->bytes,
                 0);
        if (n <= 0) {
          KillConnection(pConn);
          return TRUE;
        }
      }
      pConn->nServerBufLen -= m->bytes;
      pConn->nServerBufPos += m->bytes;
    }
  }
  else {
    // forward all available data
    n = send(pConn->sdClientSock,
             pConn->ServerStageBuf+pConn->nServerBufPos,
             pConn->nServerBufLen,
             0);
    if (n <= 0) {
      KillConnection(pConn);
      return TRUE;
    }
    pConn->nServerBufLen -= n;
    pConn->nServerBufPos += n;
  }

  if (pConn->nServerBufLen == 0) {
    pConn->nServerBufPos = 0;
  }

  return FALSE;
}


//============================================================================
static unsigned MsgLoop() {
  struct sockaddr_in name;
  int namelen;
  fd_set fds;
  int n;
  int newsock;

  // allocate sockets
  sgsdUnifiedSock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sgsdUnifiedSock == -1)
    return TRUE;

  // bind to the desired port
  memset(&name, '\0', sizeof(name));
  name.sin_family = AF_INET;
  name.sin_addr.s_addr = sguListenAddr;
  name.sin_port = htons(UNIFIEDPORT);
  if (bind(sgsdUnifiedSock, (struct sockaddr *)&name, sizeof(name)) < 0) {
    LogWrite("Can't bind to unified port.\n");
    return TRUE;
  }

  // start listening for connections
  listen(sgsdUnifiedSock, 5);

  // enter main loop
  for (;;) {
    int hisock = 0;

    // add the master socket
    FD_ZERO(&fds);
    FD_SET(sgsdUnifiedSock, &fds);
    hisock = sgsdUnifiedSock;

    // add active connections
    PROXYCONNECTIONPTR *pNext = &sgpConnHead;
    PROXYCONNECTIONPTR pConn;
    while ((pConn = *pNext) != 0) {
      if (pConn->uFlags & CF_DESTROY) {
        // destroy this connection record
        *pNext = pConn->pNext;
        free(pConn);
        continue;
      }
      else {
        FD_SET(pConn->sdClientSock,&fds);
        if (pConn->sdClientSock > hisock)
          hisock = pConn->sdClientSock;
        FD_SET(pConn->sdServerSock,&fds);
        if (pConn->sdServerSock > hisock)
          hisock = pConn->sdServerSock;
      }
      pNext = &pConn->pNext;
    }

    // if all proxied connections are closed, close down the UDP socket also
    if (!sgpConnHead && (sgsdUdpSock != -1)) {
      closesocket(sgsdUdpSock);
      sgsdUdpSock = -1;
    }

    // add the UDP socket if a session is active
    if (sgsdUdpSock != -1) {
      FD_SET(sgsdUdpSock,&fds);
      if (sgsdUdpSock > hisock)
        hisock = sgsdUdpSock;
    }

    // wait for some input
    n = select(hisock+1, &fds, 0, 0, 0);

    if (n > 0) {
      // check listening sockets first
      if (FD_ISSET(sgsdUnifiedSock, &fds)) {
        // accept connection
        namelen = sizeof(name);
        newsock = accept(sgsdUnifiedSock, (struct sockaddr *)&name, &namelen);

        LogWrite("Accepting connection from %s", inet_ntoa(name.sin_addr));

        // set up a new connection record
        if (!AddConnection(newsock, &name)) {
          closesocket(newsock);
        }
        n--;
      }

      // check for input from UDP socket
      if ((sgsdUdpSock != -1) && FD_ISSET(sgsdUdpSock,&fds)) {
        ProcessUdpData();
        n--;
      }

      // check for input from proxied connections
      PROXYCONNECTIONPTR pConn = sgpConnHead;
      while (pConn && n) {
        if (FD_ISSET(pConn->sdClientSock,&fds)) {
          ProcessClientData(pConn);
          n--;
        }

        if (FD_ISSET(pConn->sdServerSock,&fds)) {
          ProcessServerData(pConn);
          n--;
        }
        pConn = pConn->pNext;
      }
    }
  }
  return FALSE;
}


//=============================================================================
#ifdef WIN32
static void __cdecl IntHandler(int nErr) {
#else
static void IntHandler(int nErr) {
#endif

  LogWrite("INTERRUPT\n");

  Cleanup();
  exit (1);
}

//=============================================================================
#ifdef WIN32
static void __cdecl SegvHandler(int nErr) {
#else
static void SegvHandler(int nErr) {
#endif

  LogWrite("SEGMENTATION FAULT\n");

  Cleanup();
  exit (1);
}

//============================================================================
int Daemonize() {
#ifndef WIN32
int childpid, fd;

  // Don't need to detach if we were started by init
  if (getppid() == 1)
    goto done; 

  if ((childpid = fork()) < 0) {
    printf("cannot fork daemon.\n");
    return 0;
  }
  else if (childpid > 0)
    exit(0);

  if (setpgrp() == -1)
    exit(-2);

  // ignore the hangup signal
  signal(SIGHUP, SIG_IGN);

  if ((childpid = fork()) < 0) {
    printf("cannot fork second child.\n");
    return 0;
  }
  else if (childpid > 0)
    exit(0);

done:
  errno = 0;

  // we don't care when the children quit
  signal(SIGCLD, SIG_IGN);
#endif

  return TRUE;
}


//=============================================================================
static int Init() {

#ifdef WIN32
  // initialize winsock
  {
    WSADATA wsadata;
    if (WSAStartup(0x101,&wsadata)) {
      return FALSE;
    }
    if (wsadata.wVersion != 0x101) {
      return FALSE;
    }
  }
#endif

  return (Daemonize());
}

//=============================================================================
static int Cleanup() {

  LogWrite("Exiting.\n");
  LogClose();
#ifdef WIN32
  WSACleanup();
#endif
  return TRUE;
}

//=============================================================================
#ifdef WIN32
int __cdecl main(int argc, char *argv[]) {
#else
int main(int argc, char *argv[]) {
#endif

  LogOpen("bnpd.log");
  LogWrite("Startup.\n");

  signal(SIGSEGV, SegvHandler);
  signal(SIGINT, IntHandler);

  // process command-line options here
  ParseArgs(argc,argv);

  if (Init())
    MsgLoop();

  Cleanup();
  return 0;
}
